//-----------------------------------------------------------------------------
// name: VisualSine.cpp
// desc: hello sine wave, real-time
//
// author: Dylan Freedman (dfreed@stanford.edu)
//   date: fall 2017
//   uses: RtAudio by Gary Scavone
//-----------------------------------------------------------------------------
#include "RtAudio/RtAudio.h"
#include "chuck.h"
#include "../core/util_xforms.h"
#include <math.h>
#include <stdlib.h>
#include <iostream>
using namespace std;

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#define max(a,b) \
({ __typeof__ (a) _a = (a); \
    __typeof__ (b) _b = (b); \
  _a > _b ? _a : _b; })


//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void initGfx();
void idleFunc();
void displayFunc();
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );

// our datetype
#define SAMPLE float
// corresponding format for RtAudio
#define MY_FORMAT RTAUDIO_FLOAT32
// sample rate
const t_CKFLOAT MY_SRATE = 44100;
// number of channels
const t_CKINT MY_CHANNELS = 1;
// for convenience
#define MY_PIE 3.14159265358979

// width and height
long g_width = 1024;
long g_height = 720;
float globalFreq = 440.0;

float depth = 1;
float mult = 3000;

float currentMean = 0.0;
// global buffer
SAMPLE * g_buffer = NULL;
SAMPLE ** g_buffer_cache = NULL;

bool useChuck = false;
bool showRaw = true;
bool showFft = true;
bool showBackground = true;

long g_bufferSize = 8192;

unsigned int bufferFrames = 1024;
// fft buffer
SAMPLE * fft_buffer = NULL;
// window buffer
SAMPLE * window_buffer = NULL;
// log buffer
SAMPLE * log_buffer = NULL;
long log_bufferSize = 147;

SAMPLE * current_log_buffer = NULL;

SAMPLE ** log_buffer_cache = NULL;
int logBufferCacheSize = 20;

float depthPow = 1.0;

int frequencyRectangles = 200;
bool showFrequency = true;
bool collapseToOneOctave = true;
FLOAT * rectangles = NULL;

// global variables
bool g_draw_dB = false;
ChucK * the_chuck = NULL;
bool useAlpha = false;

typedef struct {
    double r;       // a fraction between 0 and 1
    double g;       // a fraction between 0 and 1
    double b;       // a fraction between 0 and 1
} rgb;

rgb goalColor;
rgb currentColor = {0, 0, 0};

rgb hsv2rgb(float h, float s, float v)
{
    double      hh, p, q, t, ff;
    long        i;
    rgb         out;

    if(s <= 0.0) {       // < is bogus, just shuts up warnings
        out.r = v;
        out.g = v;
        out.b = v;
        return out;
    }
    hh = h * 360.0;
    if(hh >= 360.0) hh = 0.0;
    hh /= 60.0;
    i = (long)hh;
    ff = hh - i;
    p = v * (1.0 - s);
    q = v * (1.0 - (s * ff));
    t = v * (1.0 - (s * (1.0 - ff)));

    switch(i) {
    case 0:
        out.r = v;
        out.g = t;
        out.b = p;
        break;
    case 1:
        out.r = q;
        out.g = v;
        out.b = p;
        break;
    case 2:
        out.r = p;
        out.g = v;
        out.b = t;
        break;

    case 3:
        out.r = p;
        out.g = q;
        out.b = v;
        break;
    case 4:
        out.r = t;
        out.g = p;
        out.b = v;
        break;
    case 5:
    default:
        out.r = v;
        out.g = p;
        out.b = q;
        break;
    }
    return out;
}

//-----------------------------------------------------------------------------
// name: initChucK()
// desc: initialize ChucK
//-----------------------------------------------------------------------------
bool initChucK()
{
  // Instantiate ChucK.
  the_chuck = new ChucK();
  // Set params.
  the_chuck->setParam("SAMPLE_RATE", MY_SRATE);
  the_chuck->setParam("INPUT_CHANNELS", MY_CHANNELS);
  the_chuck->setParam("OUTPUT_CHANNELS", MY_CHANNELS);
  // Initialize ChucK and return the status.
  return the_chuck->init();
}


//-----------------------------------------------------------------------------
// name: callme()
// desc: audio callback
//-----------------------------------------------------------------------------
int callme( void * outputBuffer, void * inputBuffer, unsigned int numFrames,
            double streamTime, RtAudioStreamStatus status, void * data )
{
    // cast!
    SAMPLE * input = (SAMPLE *)inputBuffer;
    SAMPLE * output = (SAMPLE *)outputBuffer;

    if (useChuck) {
        the_chuck->run(input, output, numFrames);
    }

    // shift
    for (int i = g_bufferSize - bufferFrames - 1; i >= 0; i--) {
        g_buffer[i] = g_buffer[i + bufferFrames];
    }
    // fill
    for( int i = 0; i < numFrames; i++ )
    {
        // assume mono
        g_buffer[i + g_bufferSize - bufferFrames] = input[i];
        // zero output
        if (!useChuck) {
            output[i] = 0;
        }
    }

    return 0;
}




//-----------------------------------------------------------------------------
// name: main()
// desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // instantiate RtAudio object
    RtAudio audio;
    // variables
    unsigned int bufferBytes = 0;

    // check for audio devices
    if( audio.getDeviceCount() < 1 )
    {
        // nopes
        cout << "no audio devices found!" << endl;
        exit( 1 );
    }

    // initialize GLUT
    glutInit( &argc, argv );
    // init gfx
    initGfx();

    // let RtAudio print messages to stderr.
    audio.showWarnings( true );

    // set input and output parameters
    RtAudio::StreamParameters iParams, oParams;
    iParams.deviceId = audio.getDefaultInputDevice();
    iParams.nChannels = MY_CHANNELS;
    iParams.firstChannel = 0;
    oParams.deviceId = audio.getDefaultOutputDevice();
    oParams.nChannels = MY_CHANNELS;
    oParams.firstChannel = 0;

    // create stream options
    RtAudio::StreamOptions options;

    // go for it
    try {
        // open a stream
        audio.openStream( &oParams, &iParams, MY_FORMAT, MY_SRATE, &bufferFrames, &callme, (void *)&bufferBytes, &options );
    }
    catch( RtError& e )
    {
        // error!
        cout << e.getMessage() << endl;
        exit( 1 );
    }

    // compute
    bufferBytes = bufferFrames * MY_CHANNELS * sizeof(SAMPLE);
    // allocate global buffer
    g_bufferSize = bufferFrames;
    g_buffer = new SAMPLE[g_bufferSize];
    fft_buffer = new SAMPLE[g_bufferSize * 2];
    window_buffer = new SAMPLE[g_bufferSize];
    log_buffer = new SAMPLE[log_bufferSize];
    current_log_buffer = new SAMPLE[log_bufferSize];

    memset( g_buffer, 0, sizeof(SAMPLE)*g_bufferSize );
    memset( current_log_buffer, 0, sizeof(SAMPLE)*log_bufferSize );
    memset( window_buffer, 0, sizeof(SAMPLE)*g_bufferSize );
    memset( fft_buffer, 0, sizeof(SAMPLE)*g_bufferSize *2 );


    log_buffer_cache = new SAMPLE*[logBufferCacheSize];
    for (int i = 0; i < logBufferCacheSize; i++) {
        log_buffer_cache[i] = new SAMPLE[log_bufferSize];
        memset(log_buffer_cache[i], 0, sizeof(SAMPLE)*log_bufferSize);
    }

    rectangles = new FLOAT[frequencyRectangles];
    memset(rectangles, 0, sizeof(FLOAT) * frequencyRectangles);

    // Make the window
    hamming(window_buffer, g_bufferSize);

    // NOTE: init ChucK (see function above)
    if( !initChucK() ) exit( 1 );

    if (!the_chuck->compileFile("driver.ck", "")) {
        cerr << "ChucK code could not be compiled.\n";
        exit(1);
    }

    the_chuck->broadcastExternalEvent("e");
    the_chuck->broadcastExternalEvent("TestEvent");

    // go for it
    try {
        // start stream
        audio.startStream();

        // let GLUT handle the current thread from here
        glutMainLoop();

        // stop the stream.
        audio.stopStream();
    }
    catch( RtError& e )
    {
        // print error message
        cout << e.getMessage() << endl;
        goto cleanup;
    }

cleanup:
    // close if open
    if( audio.isStreamOpen() )
        audio.closeStream();

    // done
    return 0;
}


void printUsage() {
    cout << "Welcome to HueNote!\n"
         << "-=-=-=-=-=-=-=-=-=-\n"
         << "Keyboard Usage:\n"
         << "  q: Quit\n"
         << "  a: Toggle display of raw signal in background\n"
         << "  s: Toggle display of circles\n"
         << "  z: Toggle fading inner circles\n"
         << "  c: Play a messed up Shephard tone\n"
         << "  x: Toggle collapsing fft to one octave\n"
         << "  b: Toggle showing colored background\n"
         << "  ?: Print this usage message\n"
         << "Mouse controls:\n"
         << "  Left : Increase the spacing between circles\n"
         << "  Right: Decrease the spacing between circles\n"
         << "Special control (only works on Mac OS X with Chrome installed):\n"
         << "  Cmd+Space c h r o m e Enter Cmd+T\n"
         << "  m a i l t o : f r e e d m a n d @ g m a i l . c o m: Emails author\n";
}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void initGfx()
{
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "VisualSine" );

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );

    // set clear color
    glClearColor( 0, 0, 0, 1 );
    // enable color material
    glEnable( GL_COLOR_MATERIAL );
    // enable depth test
    glEnable( GL_DEPTH_TEST );
    printUsage();
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    int semis = -1;
    switch( key )
    {
        case 'Q':
        case 'q':
            exit(1);
            break;
        case 'a':
            showRaw = !showRaw;
            cout << "show raw signal " << showRaw << "\n";
            break;
        case 's':
            showFft = !showFft;
            cout << "show fft circles " << showFft << "\n";
            break;
        case 'x':
            collapseToOneOctave = !collapseToOneOctave;
            cout << "collapse to one octave " << collapseToOneOctave << "\n";
            break;
        case 'c':
            useChuck = !useChuck;
            cout << "use chuck " << useChuck << "\n";
            break;
        case 'z':
            useAlpha = !useAlpha;
            cout << "alpha " << useAlpha << "\n";
            break;
        case 'b':
            showBackground = !showBackground;
            cout << "show background " << showBackground << "\n";
        case '?':
            printUsage();
            break;
    }

    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // when left mouse button is down
        if( state == GLUT_DOWN )
        {
            mult *= 1.2;
            cout << "Size mult " << mult << "\n";
        }
        else
        {
        }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        // when right mouse button down
        if( state == GLUT_DOWN )
        {
            mult *= 0.8;
            cout << "Size mult " << mult << "\n";
        }
        else
        {
        }
    }
    else
    {
    }

    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}

void interleaved_abs(FLOAT * x, long len) {
    for (int i = 0; i < len / 2; i++) {
        FLOAT re = x[i * 2];
        FLOAT im = x[i * 2 + 1];
        x[i] = sqrt(re * re + im * im);
    }
}

float getFreq(int i) {
    FLOAT freq;
    if (i == 1) {
        freq = (float)MY_SRATE / 2;
    } else {
        freq = (float)i * MY_SRATE / g_bufferSize / 2;
    }
    // Skip frequencies that are not related to pitch perception.
    if (freq <= 0 || freq >= 11000) return -1;
    return freq;
}

void emphasize_harmonic_good_things() {

}

void subtract_harmonics() {
    // Dampen the effect of harmonics.
    for (int i = g_bufferSize - 1; i >= 0; i--) {
        // Go up to the 6th harmonic.
        for (int h = 2; h <= 6; h++) {
            int index = i / h;
            float freq = getFreq(index);
            if (freq < 0) continue;
            fft_buffer[index] *= 1 - fft_buffer[i] * (1.0 / (float)h);
        }
    }
}

void calc_log_buffer() {
    for (int i = 0; i < log_bufferSize; i++) {
        log_buffer[i] = -1;
    }
    float maxCollapsed = log(MY_SRATE / 2) / log(2);
    float minCollapsed = log(1.0 * MY_SRATE / g_bufferSize / 2) / log(2);
    for (int i = 0; i < g_bufferSize; i++) {
        float freq = getFreq(i);
        if (freq <= 0) continue;

        int result;
        if (collapseToOneOctave) {
            result = (int)(fmod(log(freq) / log(2), 1) * log_bufferSize);
        } else {
            result = (int)(((log(freq) / log(2)) - minCollapsed) / (maxCollapsed - minCollapsed) * log_bufferSize);
        }
        float value = fft_buffer[i] * fft_buffer[i];
        if (log_buffer[result] < 0) {
            log_buffer[result] = value;
        } else {
            log_buffer[result] += value;
        }
    }
}

void normalizeStdDevsBelowToZero(FLOAT * x, long len, int stdDevs) {
    float total = 0.0;
    int count = 0;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        total += x[i];
        count++;
    }
    float mean = total / count;

    // Calculate std dev
    float squaredDiffTotal = 0.0;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        float diff = x[i] - mean;
        squaredDiffTotal += diff * diff;
    }
    float stdDev = squaredDiffTotal / count;

    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        if (x[i] < mean - stdDev * stdDevs) x[i] = -1.0;
    }
}

void normalizeStdDevs(FLOAT * x, long len, float stdDevMin, float stdDevMax) {
    float total = 0.0;
    int count = 0;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        total += x[i];
        count++;
    }
    float mean = total / count;

    // Calculate std dev
    float squaredDiffTotal = 0.0;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        float diff = x[i] - mean;
        squaredDiffTotal += diff * diff;
    }
    float stdDev = sqrt(squaredDiffTotal / count);

    // Calculate min and max std devs
    float minStdDev = 0.0;
    float maxStdDev = 0.0;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        float stdDevs = (x[i] - mean) / stdDev;
        stdDevs *= stdDevs;
        if (stdDevs > maxStdDev) maxStdDev = stdDevs;
        if (stdDevs < stdDevMin) {
            x[i] = -1.0;
        } else if (stdDevs < minStdDev) {
            minStdDev = stdDevs;
        }
    }

    // if (stdDevMin < minStdDev) minStdDev = stdDevMin;
    if (stdDevMax > maxStdDev) maxStdDev = stdDevMax;

    float stdDevRange = maxStdDev - minStdDev;
    // Normalize.
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        float stdDevs = (x[i] - mean) / stdDev;
        stdDevs *= stdDevs;
        x[i] = (stdDevs - minStdDev) / stdDevRange;
    }
}

void normalize(FLOAT * x, long len) {
    float min = 1.0;
    float max = 0.0;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        float value = x[i];
        if (value < min) min = value;
        if (value > max) max = value;
    }

    // Normalize.
    float range = max - min;
    for (int i = 0; i < len; i++) {
        if (x[i] < 0) continue;
        x[i] = (x[i] - min) / range;
    }
}

void circularInterpolate(FLOAT * x, long len) {
    float minValue = 0.0;
    int lastIndex, firstIndex;
    // Get last non-zero value.
    for (lastIndex = len - 1; lastIndex >= 0; lastIndex--) {
        if (x[lastIndex] < minValue) continue;
        break;
    }
    // Get first non-zero value.
    for (firstIndex = 0; firstIndex < len; firstIndex++) {
        if (x[firstIndex] < minValue) continue;
        break;
    }
    if (firstIndex >= lastIndex) {
        // Can't interpolate if there are no values.
        return;
    }
    // Interpolate the wrap-around.
    int range = firstIndex + len - lastIndex;
    float valueRange = x[firstIndex] - x[lastIndex];
    for (int i = lastIndex + 1; i < firstIndex + len; i++) {
        int effectiveIndex = i % len;
        float progress = ((float)(i - lastIndex)) / range;
        x[effectiveIndex] = x[lastIndex] + progress * valueRange;
    }

    int prevIndex = firstIndex;
    for (int i = firstIndex + 1; i <= lastIndex; i++) {
        if (x[i] < minValue) {
            continue;
        }
        // Interpolate between found values.
        float prevValue = x[prevIndex];
        int range = i - prevIndex;
        float valueRange = x[i] - prevValue;
        for (int j = prevIndex + 1; j < i; j++) {
            float progress = ((float)(j - prevIndex)) / range;
            x[j] = prevValue + progress * valueRange;
        }
        prevIndex = i;
    }
}

void shiftCache(SAMPLE ** cache, int cacheLen, SAMPLE * current, int subsize) {
    for (int i = 0; i < cacheLen - 1; i++) {
        for (int j = 0; j < subsize; j++) {
            SAMPLE prev = cache[i + 1][j];
            cache[i][j] = cache[i + 1][j];
        }
    }
    for (int j = 0; j < subsize; j++) {
        cache[cacheLen - 1][j] = current[j];
    }
}

void drawLine(double r, double g, double b, SAMPLE * data, double multiplier, long len, float offset, float alpha) {
    // color
    glColor3f( r * alpha, g * alpha, b * alpha );

    // define a starting point
    GLfloat x = -5;
    // increment
    GLfloat xinc = ::fabs(x*2 / len);

    // start primitive
    glBegin( GL_LINE_STRIP );

    // loop over buffer
    for( int i = 0; i < len; i++ )
    {
        // plot
        glVertex3f( x, multiplier*data[i], offset * 10 );
        // increment x
        x += xinc;
    }

    // end primitive
    glEnd();
}

void drawBlurredLine(double r, double g, double b, double alpha, double x1, double y1, double z1, double x2, double y2, double z2, int width) {
    for (int i = width - 1; i >= 0; i--) {
        // Set color
        glColor4f( r, g, b, alpha * (1.0 / (width - i)) );
        // line width
        glLineWidth( (width - i) );
        glBegin( GL_LINE_STRIP );

        glVertex3f( x1, y1, z1 );
        glVertex3f( x2, y2, z2 );

        glEnd();
    }
}

rgb interpColor(rgb current, rgb goal, float slew) {
    rgb result;
    result.r = current.r + (goal.r - current.r) * slew;
    result.g = current.g + (goal.g - current.g) * slew;
    result.b = current.b + (goal.b - current.b) * slew;
    return result;
}

//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // shiftCache();

    // apply rfft with window and store in fft_buffer
    memcpy(fft_buffer, g_buffer, g_bufferSize);
    apply_window(fft_buffer, window_buffer, g_bufferSize);
    rfft(fft_buffer, g_bufferSize, true);
    interleaved_abs(fft_buffer, g_bufferSize * 2);

    // Subtract the effect of harmonics.
    // subtract_harmonics();

    // Calculate the log buffer.
    calc_log_buffer();
    normalize(log_buffer, log_bufferSize);
    // normalizeStdDevsBelowToZero(log_buffer, log_bufferSize, 0.0);
    normalizeStdDevs(log_buffer, log_bufferSize, 10, 400.0);
    circularInterpolate(log_buffer, log_bufferSize);
    // normalize(log_buffer, log_bufferSize);

    // Calc mean
    float total = 0.0;
    for (int i = 0; i < g_bufferSize; i++) {
        total += fabs(g_buffer[i]);
    }
    float mean = total / g_bufferSize;
    currentMean += 0.01 * (mean - currentMean);
    for (int i = 0; i < log_bufferSize; i++) {
        log_buffer[i] *= currentMean;
    }

    for (int i = 0; i < log_bufferSize; i++) {
        current_log_buffer[i] = current_log_buffer[i] + 0.1 * (log_buffer[i] - current_log_buffer[i]);
    }
    shiftCache(log_buffer_cache, logBufferCacheSize, current_log_buffer, log_bufferSize);

    // local state
    static GLfloat zrot = 0.0f, c = 0.0f;

    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);

    float prevX = 0;
    float prevY = 0;
    float firstX = 0;
    float firstY = 0;
    float maxHue = 0.0;
    float maxHueValue = 0.0;
    float firstHue = 0;
    if (showFft) {
        for (int j = logBufferCacheSize - 1; j >= 0; j--) {
            float alpha = (float)j / (logBufferCacheSize - 1);
            float adjustedDepth = -depth + alpha * depth;
            for (int i = 0; i < log_bufferSize; i++) {
                float hue = (float)i / log_bufferSize;
                rgb color = hsv2rgb(hue, 1.0, 1.0);
                float theta = hue * 2 * MY_PIE;
                float r = (log_buffer_cache[j][i] * mult + 2) * alpha;
                if (j == logBufferCacheSize - 1 && r > maxHueValue) {
                    maxHueValue = r;
                    maxHue = hue;
                }

                // Convert to rectangular.
                float x = r * cos(theta);
                float y = r * sin(theta);
                if (i != 0) {
                    drawBlurredLine(color.r, color.g, color.b, useAlpha ? alpha : 1.0, prevX, prevY, adjustedDepth, x, y, adjustedDepth, 1);
                } else {
                    firstX = x;
                    firstY = y;
                    firstHue = hue;
                }
                prevX = x;
                prevY = y;
            }
            rgb color = hsv2rgb(firstHue, 1.0, 1.0);
            drawBlurredLine(color.r, color.g, color.b, useAlpha ? alpha : 1.0, prevX, prevY, adjustedDepth, firstX, firstY, adjustedDepth, 1);
        }
    }

    if (showRaw) {
        glColor4f( 1.0, 1.0, 1.0, 0.05 );
        float increment = (float)g_bufferSize / frequencyRectangles;
        for (int i = 0; i < frequencyRectangles; i++) {
            float total = 0.0;
            int count = 0;
            for (int j = increment * i; j < increment * (i + 1); j++) {
                total += g_buffer[j];
                count++;
            }
            float avg = total / count;
            // rectangles[i] = avg;
            if (isnan(avg)) continue;
            rectangles[i] += 0.05 * (avg - rectangles[i]);
            glRectf((((float)(i) - 0.5) / frequencyRectangles - 0.5) * 20, -rectangles[i] * 150, ((((float)i) + 0.5) / frequencyRectangles - 0.5) * 20, rectangles[i] * 150);
        }
    }

    if (showBackground) {
        goalColor = hsv2rgb(maxHue, 1, 0.2);
        currentColor = interpColor(currentColor, goalColor, 0.02);
        glClearColor(currentColor.r, currentColor.g, currentColor.b, 1.0);
    } else {
        glClearColor(0, 0, 0, 1.0);
    }

    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
}
